#   Copyright [2023] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..pubsub import PubSubManager
from ..logging import setup_logging

logging = setup_logging()

import datetime


def archive_qa(bot_output, vector_name):
    try:
        pubsub_manager = PubSubManager(vector_name, pubsub_topic="qna-to-pubsub-bq-archive")
        the_data = {"bot_output": bot_output,
                    "vector_name": vector_name,
                    "timestamp": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}
        
        pubsub_manager.publish_message(the_data)
    except Exception as e:
        logging.warning(f"Could not publish message for {vector_name} to qna-to-pubsub-bq-archive - {str(e)}")


#   Copyright [2023] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
from ..logging import setup_logging

logging = setup_logging()


def pick_vectorstore(vs_str, vector_name, embeddings):
    logging.debug('Picking vectorstore')
        
    if vs_str == 'supabase':
        from supabase import Client, create_client
        from langchain.vectorstores import SupabaseVectorStore
        from ..database.database import setup_supabase

        logging.debug(f"Initiating Supabase store: {vector_name}")
        setup_supabase(vector_name)

        # init embedding and vector store
        supabase_url = os.getenv('SUPABASE_URL')
        supabase_key = os.getenv('SUPABASE_KEY')

        logging.debug(f"Supabase URL: {supabase_url} vector_name: {vector_name}")

        supabase: Client = create_client(supabase_url, supabase_key)

        vectorstore = SupabaseVectorStore(supabase, 
                                        embeddings,
                                        table_name=vector_name,
                                        query_name=f'match_documents_{vector_name}')

        logging.debug("Chose Supabase")
    elif vs_str == 'cloudsql':
        from langchain.vectorstores.pgvector import PGVector

        logging.debug("Inititaing CloudSQL pgvector")
        #setup_cloudsql(vector_name) 

        # https://python.langchain.com/docs/modules/data_connection/vectorstores/integrations/pgvector
        CONNECTION_STRING = os.environ.get("PGVECTOR_CONNECTION_STRING")
        # postgresql://brainuser:password@10.24.0.3:5432/brain

        from ..database.database import get_vector_size
        vector_size = get_vector_size(vector_name)

        os.environ["PGVECTOR_VECTOR_SIZE"] = str(vector_size)
        vectorstore = PGVector(connection_string=CONNECTION_STRING,
            embedding_function=embeddings,
            collection_name=vector_name,
            #pre_delete_collection=True # for testing purposes
            )

        logging.debug("Chose CloudSQL")
    elif vs_str == 'alloydb':  # exact same as CloudSQL for now
        from langchain.vectorstores.pgvector import PGVector

        logging.info("Inititaing AlloyDB pgvector")
        #setup_cloudsql(vector_name) 

        # https://python.langchain.com/docs/modules/data_connection/vectorstores/integrations/pgvector
        CONNECTION_STRING = os.environ.get("ALLOYDB_CONNECTION_STRING",None)
        if CONNECTION_STRING is None:
            logging.info("Did not find ALLOYDB_CONNECTION_STRING fallback to PGVECTOR_CONNECTION_STRING")
            CONNECTION_STRING = os.environ.get("PGVECTOR_CONNECTION_STRING")
        # postgresql://brainuser:password@10.24.0.3:5432/brain

        from ..database.database import get_vector_size
        vector_size = get_vector_size(vector_name)

        os.environ["PGVECTOR_VECTOR_SIZE"] = str(vector_size)
        vectorstore = PGVector(connection_string=CONNECTION_STRING,
            embedding_function=embeddings,
            collection_name=vector_name,
            #pre_delete_collection=True # for testing purposes
            )

        logging.info("Chose AlloyDB")

    else:
        raise NotImplementedError(f'No llm implemented for {vs_str}')   

    return vectorstore


#   Copyright [2023] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..logging import setup_logging

logging = setup_logging()
import requests
import aiohttp

from .route import route_endpoint

def send_to_qa(user_input, vector_name, chat_history, stream=False, **kwargs):

    # {'stream': '', 'invoke': ''}
    endpoints = route_endpoint(vector_name)

    if stream:
        qna_endpoint = endpoints["stream"]
    else:
        qna_endpoint = endpoints["invoke"]

    # Base qna_data dictionary
    qna_data = {
        'user_input': user_input,
        'chat_history': chat_history
    }

    # Update qna_data with optional values from kwargs
    qna_data.update(kwargs)

    try:
        logging.info(f"Sending to {qna_endpoint} this data: {qna_data}")
        qna_response = requests.post(qna_endpoint, json=qna_data, stream=stream)
        qna_response.raise_for_status()

        if stream:
            # If streaming, return a generator that yields response content chunks
            def content_generator():
                for chunk in qna_response.iter_content(chunk_size=8192):
                    yield chunk
            return content_generator()
        else:
            # Otherwise, return the JSON response directly
            return qna_response.json()

    except requests.exceptions.HTTPError as err:
        logging.error(f"HTTP error occurred: {err}")
        error_message = f"There was an error processing your request. Please try again later. {str(err)}"
        if stream:
            return iter([error_message])
        else:
            return {"answer": error_message}

    except Exception as err:
        logging.error(f"Other error occurred: {str(err)}")
        error_message = f"Something went wrong. Please try again later. {str(err)}"
        if stream:
            return iter([error_message])
        else:
            return {"answer": error_message}

async def send_to_qa_async(user_input, vector_name, chat_history):

    # {'stream': '', 'invoke': ''}
    endpoints = route_endpoint(vector_name)

    qna_endpoint = endpoints['invoke']
    qna_data = {
        'user_input': user_input,
        'chat_history': chat_history,
    }
    logging.info(f"Sending to {qna_endpoint} this data: {qna_data}")
    
    async with aiohttp.ClientSession() as session:
        async with session.post(qna_endpoint, json=qna_data) as resp:
            qna_response = await resp.json()

    logging.info(f"Got back QA response: {qna_response}")
    return qna_response



#   Copyright [2023] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..logging import setup_logging

logging = setup_logging()
import base64
import json

def process_pubsub(data):

    logging.debug(f'process_pubsub: {data}')
    message_data = base64.b64decode(data['message']['data']).decode('utf-8')
    messageId = data['message'].get('messageId')
    publishTime = data['message'].get('publishTime')

    logging.debug(f"This Function was triggered by messageId {messageId} published at {publishTime}")
    # DANGER: Will trigger this dunction recursivly
    #logging.info(f"bot_help.process_pubsub message data: {message_data}")

    try:
        message_data = json.loads(message_data)
    except:
        logging.debug("Its not a json")

    if message_data:
        return message_data
    
    logging.info(f"message_data was empty")
    return ''


#   Copyright [2023] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from threading import Thread, Event
from queue import Queue
import json
import time

from .content_buffer import ContentBuffer, BufferStreamingStdOutCallbackHandler

from ..qna.parsers import parse_output

from ..logging import setup_logging

logging = setup_logging()

def start_streaming_chat(question, 
                         vector_name,
                         qna_func,
                         chat_history=[],
                         wait_time=2,
                         timeout=120, # Timeout in seconds (2 minutes)
                         **kwargs): 

    # Immediately yield to indicate the process has started.
    yield "Thinking...\n"
    # Initialize the chat
    content_buffer = ContentBuffer()
    chat_callback_handler = BufferStreamingStdOutCallbackHandler(content_buffer=content_buffer, tokens=".!?\n")

    result_queue = Queue()
    exception_queue = Queue()  # Queue for exceptions
    stop_event = Event()

    def start_chat(stop_event, result_queue, exception_queue):
        # autogen_qna(user_input, vector_name, chat_history=None, message_author=None):
        try:
            final_result = qna_func(question, vector_name, chat_history, callback=chat_callback_handler, **kwargs)
            result_queue.put(final_result)
        except Exception as e:
            exception_queue.put(e)


    chat_thread = Thread(target=start_chat, args=(stop_event, result_queue, exception_queue))
    chat_thread.start()

    start = time.time()
    first_start = start
    while not chat_callback_handler.stream_finished.is_set() and not stop_event.is_set():

        time.sleep(wait_time) # Wait for x seconds
        logging.info(f"heartbeat - {round(time.time() - start, 2)} seconds")
        # Check for exceptions and raise if any
        while not exception_queue.empty():
            raise exception_queue.get()
        
        content_to_send = content_buffer.read()

        if content_to_send:
            logging.info(f"==\n{content_to_send}")
            yield content_to_send
            content_buffer.clear()
            start = time.time() # reset timeout
        else:
            if time.time() - first_start < wait_time:
                # If the initial wait period hasn't passed yet, keep sending "..."
                yield "..."
            else:
                logging.info("No content to send")

        elapsed_time = time.time() - start
        if elapsed_time > timeout: # If the elapsed time exceeds the timeout
            logging.warning(f"Content production has timed out after {timeout} secs")
            break
    else:
        logging.info(f"Stream has ended after {round(time.time() - first_start, 2)} seconds")
        logging.info(f"Sending final full message plus sources...")
        
    
    # if  you need it to stop it elsewhere use 
    # stop_event.set()
    content_to_send = content_buffer.read()
    if content_to_send:
        logging.info(f"==\n{content_to_send}")
        yield content_to_send
        content_buffer.clear()

    # Stop the stream thread
    chat_thread.join()

    # the json object with full response in 'answer' and the 'sources' array
    final_result = result_queue.get()

    # parses out source_documents if not present etc.
    yield parse_output(final_result)


def generate_proxy_stream(stream_to_f, user_input, vector_name, chat_history, generate_f_output, **kwargs):
    def generate():
        json_buffer = ""  # Initialize an empty string buffer for JSON content
        inside_json = False  # Flag to track whether we're currently buffering JSON content

        for streaming_content in stream_to_f(user_input, vector_name, chat_history, stream=True, **kwargs):

            if isinstance(streaming_content, str):
                
                content_str = streaming_content

                if streaming_content.startswith('###JSON_START###'):
                    # Never happens?
                    logging.warning('Streaming content was a string with ###JSON_START###')
                else:
                    logging.info(f'Streaming got a string we return directly: {streaming_content}')
                    # just output the string as is
                    yield streaming_content
            else:
                # If it's a bytes object, decode it before further processing
                content_str = streaming_content.decode('utf-8')

            logging.info('Content_str: %s', content_str)

            while '###JSON_START###' in content_str:
                if '###JSON_END###' in content_str:
                    # Handle complete JSON object in a single chunk
                    start_index = content_str.index('###JSON_START###') + len('###JSON_START###')
                    end_index = content_str.index('###JSON_END###')
                    json_buffer = content_str[start_index:end_index]

                    try:
                        json_content = json.loads(json_buffer)
                        discord_output = generate_f_output(json_content)
                        to_client = f'###JSON_START###{json.dumps(discord_output)}###JSON_END###'
                        logging.info(f"Streaming JSON to_client:\n{to_client}")
                        yield to_client.encode('utf-8')
                    except json.JSONDecodeError as e:
                        logging.error(f"JSON decode error: {e}")
                    
                    # Prepare for next JSON object, if any
                    content_str = content_str[end_index + len('###JSON_END###'):]
                    json_buffer = ""

                else:
                    # Start JSON buffering if END marker is not in the same chunk
                    start_index = content_str.index('###JSON_START###') + len('###JSON_START###')
                    json_buffer = content_str[start_index:]
                    inside_json = True
                    break  # Exit while loop; rest will be handled by the next chunk

            if '###JSON_END###' in content_str and inside_json:
                # Handle case where END marker is in the current chunk
                end_index = content_str.index('###JSON_END###')
                json_buffer += content_str[:end_index]
                try:
                    json_content = json.loads(json_buffer)
                    discord_output = generate_f_output(json_content)
                    to_client = f'###JSON_START###{json.dumps(discord_output)}###JSON_END###'
                    logging.info(f"Streaming JSON to_client:\n{to_client}")
                    yield to_client.encode('utf-8')
                except json.JSONDecodeError as e:
                    logging.error(f"JSON decode error: {e}")
                
                content_str = content_str[end_index + len('###JSON_END###'):]
                json_buffer = ""
                inside_json = False

            if not inside_json and content_str:
                # Yield non-JSON content as it is
                logging.info(f"Streaming to client: {content_str}")
                yield content_str.encode('utf-8')

    return generate





from .chat_history import extract_chat_history
from .dispatch_to_qa import send_to_qa, send_to_qa_async
from .pubsub import process_pubsub
from .special_commands import handle_special_commands, app_to_store, handle_files
from .flask import register_qna_routes, create_app


#   Copyright [2023] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
def document_to_dict(document):
    return {
        "page_content": document.page_content,
        "metadata": document.metadata
    }

def parse_output(bot_output):
    if 'source_documents' in bot_output:
        bot_output['source_documents'] = [document_to_dict(doc) for doc in bot_output['source_documents']]
    if bot_output.get("answer", None) is None or bot_output.get("answer") == "":
        bot_output['answer'] = "(No text was returned)"
    return bot_output


#   Copyright [2023] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
import requests
import socket
import logging

def get_env_project_id():
    """
    Attempts to retrieve the project ID from environment variables.

    Returns:
        str or None: The project ID if found in environment variables, None otherwise.
    """
    return os.environ.get('GCP_PROJECT') or os.environ.get('GOOGLE_CLOUD_PROJECT')

def is_running_on_gcp():
    """
    Checks if the code is running on a Google Cloud Platform instance.

    Returns:
        bool: True if running on GCP, False otherwise.
    """
    try:
        # Google Cloud instances can reach the metadata server
        socket.setdefaulttimeout(0.1)
        socket.socket().connect(('metadata.google.internal', 80))
        return True
    except (socket.timeout, socket.error):
        return False

def get_service_account_email():
    service_email = os.getenv("GCP_DEFAULT_SERVICE_EMAIL")
    if service_email:
        return service_email
    service_email = get_metadata('instance/service-accounts/default/email')
    os.environ['GCP_DEFAULT_SERVICE_EMAIL'] = service_email
    return service_email

def get_gcp_project():
    project_id = get_env_project_id()
    if project_id:
        return project_id
    
    project_id = get_metadata('project/project-id')
    if project_id:
        os.environ["GCP_PROJECT"] = project_id 

    logging.warning("GCP Project ID not found. Ensure you are running on GCP or have the GCP_PROJECT environment variable set.")
    return None


def get_region():
    # The "instance/zone" metadata includes the region as part of the zone name
    region = os.getenv("GCP_REGION")
    if region:
        return region
    
    zone = get_metadata('instance/zone')
    if zone:
        # Split by '/' and take the last part, then split by '-' and take all but the last part
        region_zone = zone.split('/')[-1]
        region = '-'.join(region_zone.split('-')[:-1])
        os.environ["GCP_REGION"] = region
        return region
    return None

def get_metadata(stem):
    if not is_running_on_gcp():
        logging.warning("Not running on GCP, skipping metadata server fetch. For local testing set via env var instead e.g. GCP_PROJECT")
        return None
    
    metadata_server_url = f'http://metadata.google.internal/computeMetadata/v1/{stem}'

    headers = {'Metadata-Flavor': 'Google'}

    response = requests.get(metadata_server_url, headers=headers)

    if response.status_code == 200:
        return response.text
    else:
        print(f"Request failed with status code {response.status_code}")
        return None
